/*****************************************************************************
 * Date: 20100105
 * Author: Mario Roessler, DH5YM
 * License: GNU General Public License
 * Description: DCF77 Decoder
 * file: dcf77d.c
 *
 *
 *****************************************************************************/
#include <stdint.h>
#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include "uart.h"
#include "dcf77d.h"


//initialize timer and variables for dcf77 clock decoder
void dcf77_init(void)
{
  //initialize input pin
  DDR(DCF77_PORT) &= ~(_BV(DCF77_PIN));  //serial data input from DCF77
  PORT(DCF77_PORT) &= ~(_BV(DCF77_PIN)); //switch off the internal pullup

  #ifdef debugled
  DDRD |= ((1<<PD5) | (1<<PD6)); //Init Debug Led's
  #endif

  //init timer 0 (not finished)
  TCCR0B = ( (1 << CS00) | (1 << CS01) ); // Teiler: 64
  OCR0A = 249;       //1 tick every ms because 0.001s/(1/16e6Hz)/64

  TCCR0A |= (1<<WGM01);
  TCCR0A &= ~(1<<WGM00);
  TCCR0B &= ~(1<<WGM02);  //clear timer on compare match

  TIMSK0 = ( 1 << OCIE0A );         // Overflow Interrupt einschalten
  
  #ifdef dcf77debug
  uart_puts_P("initialized...\n\r");
  #endif

  return;
}


//enable Int0 Pinchange interrupt on rising edge
void int0pci_init(void)
{
  EICRA |= ((1<<ISC00) | (1<<ISC01)); //enable INT0 pin change interrupt on rising edge
  EIMSK |= (1<<INT0); //switch int0 on
  return;
}

//switch Int0 Pinchange interrupt off
void int0pci_off(void)
{
  EIMSK &= ~(1<<INT0);
  return;
}

void dcf77_demodulate(void)
{
       uint8_t arrayindex=0;  //index for adressing the correct byte for collecting the bits

       arrayindex = validbits/8;

       if(dcfones>70) {  //bit-start detected
	     #ifdef debugled
         PORTD |= (1<<PD6);
		 #endif
		 dcfones=0;
		 dcfzeros=0;
		 bitone++;
		 zerocount=0;
	   }

	   if(dcfzeros>100) { //gap detected, the carrier high ended
         #ifdef debugled
		 PORTD &= ~(1<<PD6);
		 #endif
		 dcfones=0;
		 dcfzeros=0;
		 zerocount++;
		 if(bitone==1) {   //one times 100ms high => is a zero
		   #ifdef dcf77debug
           uart_putc('0');
		   #endif
		   validbits++;
		 } else if(bitone==2) {  //two times 100ms high => is a one
		   #ifdef dcf77debug
           uart_putc('1');
		   #endif
		   datagram[arrayindex] |= (0x80 >> (validbits % 8));
		   validbits++;
		 }
		 bitone=0;
	   }

	   if(zerocount>16){  //missed second impulse in second 59, next is start of a minute
	      zerocount=0;
          #ifdef dcf77debug
		  uart_puts_P("\n\r ");
		  #endif
		  if(validbits==59)  //just if all bits have been received
		  {
		    for(uint8_t i=0;i<9;i++)  //save last dcf77 datagram
            {
               datagrambuf[i]=datagram[i];
            }
			datagrambuf[7] |= 0x01; //set last bit of buffer to indicate that this is a new one
		    #ifdef dcf77debug
			dcf77_printarray();  //print array for debug reason
			#endif
			testone=dcfones; //save current count of ones in the memory to get the next edge but only if we received correct number of bits
			
			/**/
			int0pci_init(); //enable Int0 pinchange interrupt

			uart_puts_P("\n\r");
            /**/			

			dcfones=1;
          }
		  for(uint8_t i=0;i<9;i++)  //clean the buffer
          {
			 datagram[i]=0x00;
          }
		  dcfones=0;
		  validbits=0; //reset datagram bit counter 
	   }
}

//debug function for printing the collected bit array
void dcf77_printarray(void)
{
  uint8_t i=0;
  uint8_t arrayindex=0;  //index for adressing the correct byte for collecting the bits
  uint8_t bit=0;  //bit buffer for printing

  i=0;
  uart_puts_P("\r[");
  
  while (i<59)
  {
    arrayindex = i/8;
    bit = datagrambuf[arrayindex] << (i % 8);

	if((bit & 0x80) == 0x80)
	{
       uart_putc('1');
	} else
	{
       uart_putc('0');
	}
	i++;
  }
  uart_puts_P("]\n\r ");
  return;
}

//function for getting the time/date out of the buffered dcf77 date and checking if it is valid
void dcf77_decode(void)
{
  uint8_t xtable[8]={1,2,4,8,10,20,40,80}; //table for decoding the data
  uint8_t paritycheck=0;
  uint8_t dminute=0,dhour=0,dday=0,ddow=0,dmonth=0;
  uint16_t dyear=0;

  #ifdef dcf77debug
  char buf[6];
  #endif

  //check if the current buffer is a new one (last bit in 8th octet)
  if(datagrambuf[7] & 0x01)
  {
    datagrambuf[7] &= ~0x01; //reset "new datagram" bit to prevent the datagram to be processed twice
    //decode because its new
	//check for the parity bits first, if all 3 are correct then decode the rest... otherwise skip
	if(datagrambuf[0] & 0x80) //first bit is always 0, if not => invalid data
	{
	  return;
    }
	if(getbitfrombuf(20)==0) //if start of time is not set => invalid data
	{
      return;
	}
    paritycheck = (getbitfrombuf(21)+getbitfrombuf(22)+getbitfrombuf(23)+getbitfrombuf(24)+getbitfrombuf(25)+getbitfrombuf(26)+getbitfrombuf(27)+getbitfrombuf(28)) % 2;
	if(paritycheck) //parity bit of minute wrong => invalid
	{
      return;
	}
	paritycheck = (getbitfrombuf(29)+getbitfrombuf(30)+getbitfrombuf(31)+getbitfrombuf(32)+getbitfrombuf(33)+getbitfrombuf(34)+getbitfrombuf(35)) % 2;
	if(paritycheck) //parity bit of hours wrong => invalid
	{
      return;
	}
    for(uint8_t i=36;i<59;i++)  //parity of the rest bits
	{
       paritycheck += getbitfrombuf(i);
	}
	if (paritycheck % 2) //if parity of rest is wrong => invalid
	{
       return;
	}

    #ifdef dcf77debug
	uart_puts_P("valid Data => ");
    #endif

    //decode minutes (bit 21-27)
	for(uint8_t i=0;i<7;i++)
    {
      dminute += (xtable[i]*getbitfrombuf(21+i));
	}

	//decode hours (bit 29-34)
    for(uint8_t i=0;i<6;i++)
    {
      dhour += (xtable[i]*getbitfrombuf(29+i));
	}

	//decode day (bit 36-41)
    for(uint8_t i=0;i<6;i++)
    {
      dday += (xtable[i]*getbitfrombuf(36+i));
	}

	//decode day of week (bit 42-44)
    for(uint8_t i=0;i<3;i++)
    {
      ddow += (xtable[i]*getbitfrombuf(42+i));
	}

	//decode month (bit 45-49)
    for(uint8_t i=0;i<5;i++)
    {
      dmonth += (xtable[i]*getbitfrombuf(45+i));
	}

	//decode year (bit 50-57)
    for(uint8_t i=0;i<8;i++)
    {
      dyear += (xtable[i]*getbitfrombuf(50+i));
	}
	dyear += 2000; //year offset is currently 2000 (for quite some time now)

    #ifdef dcf77debug
    //debug, print the decoded timestamp
    itoa(dhour,buf,10);
	uart_puts(buf);
    uart_putc(',');
	itoa(dminute,buf,10);
	uart_puts(buf);
    uart_putc(',');
    itoa(ddow,buf,10);
	uart_puts(buf);
    uart_putc(',');
	itoa(dday,buf,10);
	uart_puts(buf);
    uart_putc(',');
	itoa(dmonth,buf,10);
	uart_puts(buf);
    uart_putc(',');
	ltoa(dyear,buf,10);
	uart_puts(buf);

    uart_puts_P("\n\r ");
    #endif

	//you might want to set your clock here ?
    minute=dminute-1;  //because clock increases the minute counter after syncronisation
	hour=dhour;
	day=dday;
	month=dmonth;
	year=dyear;
	dcf77sync=0xFF; //reset value

  }
  //else nothing to decode
  return;
}


//get one specific bit from the buffer
uint8_t getbitfrombuf(uint8_t bitnum)
{
  uint8_t bit=0;
  uint8_t arrayindex=bitnum/8;
  uint8_t bitindex=bitnum % 8;

  bit = (datagrambuf[arrayindex] >> (7-bitindex)) & 0x01;

  return bit;
}


//clock function
void clock(void)
{
	if(second>59)
	{
      minute++;
	  second = 0;
	  if(dcf77sync!=0xFE) //increment last sync counter that provides how many minutes ago the clock was last syncronized (max 254, 255 is reset)
	  {
	    dcf77sync++;
      }
	  if(minute>59)
	  {
        hour++;
		minute=0;
		if(hour>23)
		{
          day++;
		  hour = 0;
		  if(month==2) //exceptions for february
		  {
            if(year % 4) //switching year
			{
              if(day==28)
			  {
                month++;
				day=1;
			  }
			} 
			else if(day==29)
			{
              month++;
			  day=1;
			}
		  } 
		  else if((month==1)+(month==3)+(month==5)+(month==7)+(month==8)+(month==10)+(month==12))
		  {
            if(day==31)
			{
              month++;
			  day=1;
			}
		  } 
		  else if(day==30)
		  {
            month++;
			day=1;
		  }
		  if(month>12)
		  {
            month=1;
			year++;
		  }
		}
	  }
  }

  #ifdef debugclock
  if(second!=lastsecond)
  {
    printclock();
	lastsecond=second;
  }
  #endif

  return;
}

//function for printing current clock status
void printclock(void)
{
  char buf[6];
  
  itoa(hour,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(minute,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(second,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(day,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(month,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(year,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(dcf77sync,buf,10);
  uart_puts(buf);
  uart_putc(',');
  itoa(msoffset,buf,10);
  uart_puts(buf);
  uart_puts_P("     \r"); 

  return;
}

//Timer0 COMPA ISR for sampling dcf receiver input
ISR( TIMER0_COMPA_vect )
{
	  //TCNT0=0;  //reset timer0 counter
	  dcfsample = dcfsample<<1;  //shift one bit left in the FIFO
	  dcfsample|=(PIN(DCF77_PORT)>>DCF77_PIN & 0x01); //add the current status of the dcf77 input

	  milisecond++; //increment milisecond counter
	  if(milisecond==1000) //one second has passed
	  {
        milisecond=0;
		second++;
	  }
	  if((dcfsample&0x01)==1)  //LED for debug
	   {
	      #ifdef debugled
	      PORTD |= (1<<PD5);
		  #endif
		  dcfones++;
	   }
	   else
	   {
	      #ifdef debugled
          PORTD &= ~(1<<PD5);
		  #endif
		  dcfzeros++;
	   }

	  return;
}

//INT0 pinchange interrupt service routine
ISR( INT0_vect )
{
  int0pci_off(); //switch off

  msoffset=milisecond; //save old value of ms offset
  milisecond=0;
  second=60;

  return;
}
